from email.policy import default
import click
from typing import Any, Union
from pathlib import Path
from pyfiglet import Figlet
from rich.panel import Panel
from rich import print
from jais.__init__ import NAME, VERSION, DESCRIPTION, ROOT_DIR
from jais.utils import load_default_configs

# Load project info
CONTEXT_SETTINGS = dict(auto_envvar_prefix='COMPLEX')

# ----------------------------------------> CLI ENTRY and INITIAL SETTINGS :
# Check environment variables


def check_required_settings() -> None:
    from jais.utils.fileloader import load_json
    JAIS_CNF = load_json(ROOT_DIR/'configs/jais_settings.json')
    if 'JAIS_CWD' not in JAIS_CNF.keys():
        _msg = ("JAIS's current working directory path is not set. "
                "This is required if `[bold].yaml[/bold]` configuration files "
                "contains `[bold]!cwd[/bold]` tag which renders as current working "
                "directory path."
                "\nRun [yellow]jais set-cwd /path/to/project/folder[/yellow] "
                "command to set this path."
                )
        print(Panel.fit(_msg, border_style="red"))
    else:
        click.secho("OK", fg='green')

# ENTRY POINT


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        # * BANNER
        # Find more fonts here: http://www.figlet.org/examples.html
        f = Figlet(font='smslant')
        banner = ' '.join(NAME)
        # banner = f"..._ {banner} _..."
        click.secho(f"{f.renderText(banner)}\nv{VERSION}", fg='yellow')
        # Check required settings
        check_required_settings()
        # CLI message
        click.echo(f"""Welcome to {click.style(DESCRIPTION, fg='yellow')} CLI.

    Type `{click.style(f"{NAME.lower()} --help", fg='yellow')}` for usage details
    """)
        click.echo(ctx.get_help())

    else:
        click.secho(f"\n[Running {ctx.invoked_subcommand}]...", fg='cyan')


@cli.command()
@click.argument('key')
@click.argument('value')
@click.option('--is_path',
              default=False,
              type=bool,
              is_flag=True,
              help="Use to convert relative path to absolute path if the `value` argument is a path.")
def add_item(key: Union[int, tuple, str], value: Any, is_path: bool):
    """Add key:value pair/item to jais_settings.json"""
    if is_path:
        value = str(Path(value).resolve())
    from jais.utils.fileloader import load_json, save_json
    jais_settings_path = ROOT_DIR/'configs/jais_settings.json'
    # Load settings from file or create empty dict
    JAIS_CNF = load_json(
        jais_settings_path) if jais_settings_path.exists() else {}
    # Add Current Working Directory path to settings
    JAIS_CNF[key] = value
    # Save back
    save_json(JAIS_CNF, jais_settings_path)
    print(f"Added [bold]{key}:{value}[/bold]")


@cli.command()
@click.argument('path', type=click.Path(exists=True))
def set_cwd(path: Union[str, Path]):
    """Add Current Working Directory path to settings"""
    add_item('JAIS_CWD', path, is_path=True)

# ----------------------------------------> ADDITIONAL INFORMATION :


@cli.command()
def default_configs():
    """Show default configuration settings from jais/configs/default.yaml"""
    CNF, _ = load_default_configs()
    print("Default package config file is present @", end=' ')
    print(f"`[yellow]{ROOT_DIR}/configs/default.yaml[/yellow]`")
    print("\nThis file contains the following settings:")
    print(CNF)

# ----------------------------------------> RUN EXAMPLES :


@cli.command()
@click.option(
    '-t',
    '--example_task_name',
    type=click.Choice(['cifar10', 'cifar100'], case_sensitive=False),
    default=None,
    required=True,
    show_default=None,
    help='name of example task out of the given choices.'
)
def run_example(example_task_name):
    """Run example training tasks"""
    import subprocess
    if example_task_name == 'cifar10':
        subprocess.run(['python', f"{ROOT_DIR}/examples/cifar10.py"])


if __name__ == '__main__':
    cli()



"""Data Visualization Module"""
from rich import print
from typing import Union, List
from numpy import ndarray
from torch import Tensor
import matplotlib.pyplot as plt

def show_batch(inputs: Union[ndarray, Tensor, List[ndarray]],
               targets: Union[ndarray, Tensor, List[ndarray]],
               outputs: Union[None, ndarray, Tensor, List[ndarray]] = None,
               n_cols: int = None, 
               title_fontsize: int = 20) -> None:
    """Display images in a grid

    Args:
        inputs: sample images
        targets: ground truth
        outputs: predictions
        n_cols: Number of columns to plot in the figure
        title_fontsize: font size of the main plot's title
    """
    n_inputs_dims = len(inputs.shape)

    if n_inputs_dims == 3:
        batch_size, height, width = inputs.shape
        n_channels = 1
    elif n_inputs_dims == 4:
        batch_size, n_channels, height, width = inputs.shape
    else:
        raise NotImplementedError("3D or 4D inputs supported only.")

    if n_cols is None:
        if batch_size <= 8:
            n_cols, n_rows = batch_size, 1
        else:
            n_cols = 8
            n_rows = batch_size // n_cols
    else:
        if batch_size % n_cols != 0:
            n_rows = batch_size // n_cols
            msg = f"Displaying {n_rows * n_cols} images out of {batch_size} "
            msg += f"because full grid matrix is not possible with n_cols={n_cols}."
            print(msg)
        else:
            n_rows = batch_size // n_cols

    fig, axes = plt.subplots(n_rows, n_cols, figsize=(n_cols*2, (n_rows*2)+1))
    for i, ax in enumerate(axes.flat):
        sample_image = inputs[i]
        ax.imshow(sample_image, cmap='gray' if n_channels == 1 else None)
        true_label = targets[i]
        pred_label = true_label if outputs is None else outputs[i]
        color = "green" if true_label == pred_label else "red"
        if outputs is None:
            ax.set_title(pred_label, color=color, fontdict={
                'fontsize': title_fontsize})
        else:
            ax.text(0.40, 1.1, pred_label, fontsize=20, color=color, ha='right',
                    verticalalignment='center', transform=ax.transAxes)
            ax.text(0.55, 1.1, " |", fontsize=20, color='white', ha='right',
                    verticalalignment='center', transform=ax.transAxes)
            ax.text(0.60, 1.1, true_label, fontsize=20, color='g', ha='left',
                    verticalalignment='center', transform=ax.transAxes)
        ax.axis('off')
    plt.show()



import os
import time
import logging
from typing import Optional, Union
from pathlib import Path
from jais.__init__ import ROOT_DIR


__all__ = [
    'get_logger', 'manage_log_files'
]

def get_logger(name: str,
               logs_dir: Optional[Union[str, Path]] = None,
               log_filename: Optional[Union[str, Path]] = None,
               logs_conf_filepath: Union[str, Path] = None,
               level: Optional[str] = None,
               keep_n_recent_logs: int = 5,
               rich_logger: bool = True,
               colored_logger: bool = False) -> logging.Logger:
    """Create a colored logger with two handlers (stream and file logging)
        from logs.conf

    Args:
        name: name of the logger.
        logs_dir: folder path where logs will be saved. 
            [Default is to save in the current working directory]
        log_filename: Name of the logs file (with extension `.log`).
            [Default name is `logs@<current time>.logs`]
        logs_conf_filepath: Logs configuration filepath. 
            [Default is jais/configs/logs.conf]
        level: override JAISLogger.StreamHandler logging level: 
            ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'].
            Default is loaded from the file.
        keep_n_recent_logs: Number of recent log files to keep. 
            New files will overwrite old ones. [Default = 5]
        rich_logger: Use `RichHandler` logger for colored logs
        colored_logger: Use `coloredlogs` logger for colored logs.
            Mutually exclusive to rich_logger.
    Returns:
        Colored logger instance with settings loaded from `configs/logs.conf`
    """
    if (rich_logger == True) and (colored_logger == True):
        _errmsg = "Both `rich_logger` and `colored_logger` cannot be\
             True at the same time."
        raise ValueError(_errmsg)

    import configparser  # To read logs.conf file
    import logging
    import logging.config
    if logs_conf_filepath is None:
        logs_conf_filepath = ROOT_DIR/"configs/logs.conf"
    logs_conf = configparser.RawConfigParser()
    logs_conf.read(logs_conf_filepath)
    if f'logger_{name}' in logs_conf.keys():
        default_level = logs_conf[f'logger_{name}']['level']
    else:
        default_level = 'DEBUG'
    if logs_dir is None:
        logs_dir = Path.cwd()/f"{name}_logs"
        logs_dir.mkdir(parents=True, exist_ok=True)
    else:
        logs_dir = Path(logs_dir)
    if log_filename is None:       
        log_filename = Path(logs_dir)/f"log@{time.time()}.logs"
    logging.config.fileConfig(logs_conf_filepath,
                              defaults={'logfilepath': logs_dir/log_filename},
                              disable_existing_loggers=True)
    # Create logger
    logger = logging.getLogger(name)
    # Override logging level
    if level:
        logger.setLevel(level)
        default_level = level

    if rich_logger:
        # Existing StreamHandler needs to be removed, if present.
        existing_handlers = [
            handler for handler in logger.handlers
            if isinstance(handler, logging.FileHandler)
        ]
        logger.handlers.clear()
        logger.handlers = existing_handlers
        # Add RichHandler
        from rich.logging import RichHandler
        ch = RichHandler(level=default_level,
                         show_time=False,
                         rich_tracebacks=True)
        ch.setLevel(default_level)
        formatter = logging.Formatter('%(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    elif colored_logger:
        try:
            import coloredlogs
        except ModuleNotFoundError:
            print('`coloredlogs` not installed. Use `pip install coloredlogs`')
        # StreamLogger format settings for coloredlogs
        stream_fmt = '[%(programname)s - %(levelname)s] %(message)s'
        # Set colored logs
        coloredlogs.install(level=default_level, fmt=stream_fmt, logger=logger)

    manage_log_files(logs_dir=logs_dir, 
                     keep_n_recent_logs=keep_n_recent_logs,
                     file_ext='.log')
    return logger


def manage_log_files(logs_dir: Union[str, Path], 
                     keep_n_recent_logs: int = 5, 
                     file_ext: str = '.log'):
    """Log files rotation handler"""
    # Get log files paths
    log_filespaths = list(Path(logs_dir).glob(f"*{file_ext}"))
    # Function to split the timestamp from filepath
    def get_time_from_filename(x):
        try:
            return float(x.stem.split('@')[-1])
        except ValueError:
            return None
            
    # Sort timestamps
    timestamps = sorted(
        filter(
            lambda x: False if x is None else True, 
            list(map(get_time_from_filename, log_filespaths))
        ),
        reverse=True)
    # Keep only n recent files
    timestamps = timestamps[ : keep_n_recent_logs]
    # Remove old files
    for fp in log_filespaths:
        if get_time_from_filename(fp) not in timestamps:
            os.remove(fp)



"""
* Sourse of SuperLoss, LabelSmoothingCrossEntropy, and 
    LabelSmoothingCrossEntropyWithSuperLoss):
    https://github.com/Rodger-Huang/SYSU-HCP-at-ImageCLEF-VQA-Med-2021/blob/main/utils.py
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from scipy.special import lambertw

__all__ = [
    'SuperLoss', 
    'LabelSmoothingCE',
    'LabelSmoothingCEWithSuperLoss'
]

class SuperLoss(nn.Module):
    def __init__(self, C=330, lam=0.25, device=None):
        super(SuperLoss, self).__init__()
        self.tau = torch.log(torch.FloatTensor([C]).to(device))
        self.lam = lam  # set to 1 for CIFAR10 and 0.25 for CIFAR100
        self.device = device

    def forward(self, l_i):
        l_i_detach = l_i.detach()
        # self.tau = 0.9 * self.tau + 0.1 * l_i_detach
        sigma = self.sigma(l_i_detach)
        loss = (l_i - self.tau) * sigma + self.lam * torch.log(sigma)**2
        loss = loss.mean()
        return loss

    def sigma(self, l_i):
        x = -2 / torch.exp(torch.ones_like(l_i)).to(self.device)
        y = 0.5 * torch.max(x, (l_i - self.tau) / self.lam)
        y = y.cpu().numpy()
        sigma = np.exp(-lambertw(y))
        sigma = sigma.real.astype(np.float32)
        sigma = torch.from_numpy(sigma).to(self.device)
        return sigma

class LabelSmoothingCE(nn.Module):
    def __init__(self, eps=0.1, reduction='mean', classes=330):
        super(LabelSmoothingCE, self).__init__()
        self.eps = eps
        self.reduction = reduction

    def forward(self, output, target):
        B, c = output.size()
        log_preds = F.log_softmax(output, dim=-1)
        if self.reduction == 'sum':
            loss = -log_preds.sum()
        else:
            loss = -log_preds.sum(dim=-1)
            if self.reduction == 'mean':
                loss = loss.mean()
        return loss * self.eps / c + (1 - self.eps) * F.nll_loss(log_preds, target, reduction=self.reduction)
        

class LabelSmoothingCEWithSuperLoss(nn.Module):
    def __init__(self, eps=0.1, reduction='mean', classes=330, device=None):
        super(LabelSmoothingCEWithSuperLoss, self).__init__()
        self.eps = eps
        self.reduction = reduction
        self.super_loss = SuperLoss(C=classes, device=device)
        self.device = device

    def forward(self, output, target):
        B, c = output.size()
        log_preds = F.log_softmax(output, dim=-1)
        if self.reduction == 'sum':
            loss = -log_preds.sum()
        else:
            loss = -log_preds.sum(dim=-1)
            if self.reduction == 'mean':
                loss = loss.mean()

        # l_i = (-log_preds.sum(dim=-1)) * self.eps / c + (1 - self.eps) * F.nll_loss(log_preds, target, reduction='none')
        # return self.super_loss(l_i)
        loss_cls = loss * self.eps / c + (1 - self.eps) * self.super_loss(F.nll_loss(log_preds, target, reduction='none'))
        return loss_cls


import pytest
import subprocess
from jais.utils.fileloader import load_config
from jais.__init__ import ROOT_DIR, PROJ_DIR

# * this function loads the config file for the tests.
@pytest.fixture
def CNF():
    # Set CWD path
    subprocess.run(['jais', 'set-cwd', PROJ_DIR])
    return load_config(ROOT_DIR/"configs/default.yaml")

def test_config(CNF):
    """Check if the config file is loaded correctly."""
    # Check project name
    assert CNF.project_name == "jais"
    assert isinstance(CNF.seed, int)

def test_logging(CNF):
    """Check logging settings"""
    import configparser
    logs_conf = configparser.RawConfigParser()
    logs_conf.read(ROOT_DIR/"configs/logs.conf")
    loggers = logs_conf['loggers']['keys'].split(',')
    assert (CNF.project_name in loggers[-1].lower()) == True

def test_root_dir(CNF):
    assert str(CNF.paths.root_dir) == str(ROOT_DIR)


"""The `mixup_data` and `mixup_criterion` functions are copied from 
https://github.com/facebookresearch/mixup-cifar10/blob/main/train.py
"""

import torch
import numpy as np

__all__ = ['mixup_data', 'mixup_criterion']


def mixup_data(x, y, alpha, device):
    '''Returns mixed inputs, pairs of targets, and lambda'''
    if alpha > 0:
        lam = np.random.beta(alpha, alpha)
    else:
        lam = 1

    batch_size = x.size()[0]
    index = torch.randperm(batch_size).to(device)

    mixed_x = lam * x + (1 - lam) * x[index, :]
    y_a, y_b = y, y[index]
    return mixed_x, y_a, y_b, lam


def mixup_criterion(criterion, pred, y_a, y_b, lam):
    return lam * criterion(pred, y_a) + (1 - lam) * criterion(pred, y_b)


import os
import time
import logging
from typing import Tuple
from easydict import EasyDict as edict
from .log import get_logger
from .fileloader import load_yaml
from jais.__init__ import ROOT_DIR

def get_device():
    """Get torch device instance and available GPU IDs"""
    from torch.cuda import device_count
    from torch import device
    cuda_ids = [0] if device_count() == 1 else list(range(device_count()))
    return device(f"cuda:{cuda_ids[0]}"), cuda_ids


def install_rich(verbose: bool = False):
    """Enable Rich to override Python """
    from rich import pretty, traceback
    from rich import print
    import click
    pretty.install()
    # If you are working with a framework (click, django etc),
    # you may only be interested in seeing the code from your own
    # application within the traceback. You can exclude framework
    # code by setting the suppress argument on Traceback, install,
    # and Console.print_exception, which should be a list of modules
    # or str paths.
    traceback.install(show_locals=False, suppress=[click])
    if verbose:
        print("[cyan]Rich set to override print and tracebacks.")


def load_default_configs() -> Tuple[edict, logging.Logger]:
    """Load jais package configuration settings and logger"""
    install_rich(verbose=False)
    # Load configurations
    CNF = load_yaml(f"{ROOT_DIR}/configs/default.yaml")

    # Set logger settings
    LOG_FILENAME = os.getenv('JAIS_LOG_FILENAME')
    if LOG_FILENAME is None:
        LOG_FILENAME = f"{CNF.log.filename_prefix}@{time.time()}.log"
        os.environ['JAIS_LOG_FILENAME'] = LOG_FILENAME

    LOG = get_logger(name=CNF.log.name, 
                     logs_dir=CNF.paths.logs_dir,
                     log_filename=LOG_FILENAME,
                     logs_conf_filepath=f"{ROOT_DIR}/configs/logs.conf",
                     keep_n_recent_logs=CNF.log.keep_n_recent_logs
                     )
    return CNF, LOG


import time
import torch
import torchmetrics
import torchvision
import torchvision.transforms as transforms
import torch.optim as optim
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
from torchsummary import summary
from pathlib import Path
from jais.utils import load_default_configs, get_device
from jais.visualize import show_batch
from jais.models import Net
from jais.train import *


RANK = 0
WORLD_SIZE = 1
NUM_EPOCHS = 10
MIXUP_ALPHA = 0.2
LABEL_SMOOTH = True
SUPERLOSS = True

os.environ['MASTER_ADDR'] = 'localhost'
os.environ['MASTER_PORT'] = '12356'

CNF, LOG = load_default_configs()
LOG.info("Running CIFAR10 experiment...")

training_logs_filename = f"{Path(__file__).stem}@{time.time()}"
device, cuda_ids = get_device()
dist.init_process_group('nccl', rank=RANK, world_size=WORLD_SIZE)
LOG.info(f"device = {device}")


train_tfms = transforms.Compose([
    transforms.AutoAugment(),
    transforms.ToTensor(),
    transforms.Normalize(CNF.data.mean, CNF.data.std)
])
val_tfms = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(CNF.data.mean, CNF.data.std)
])

# DATASET
classes = ('plane', 'car', 'bird', 'cat', 'deer',
           'dog', 'frog', 'horse', 'ship', 'truck')
n_classes = len(classes)

trainset = torchvision.datasets.CIFAR10(root=CNF.paths.data_dir, train=True,
                                        download=True, transform=train_tfms)
trainloader = torch.utils.data.DataLoader(trainset, batch_size=CNF.bs,
                                          shuffle=True, num_workers=2)
testset = torchvision.datasets.CIFAR10(root=CNF.paths.data_dir, train=False,
                                       download=True, transform=val_tfms)
testloader = torch.utils.data.DataLoader(testset, batch_size=CNF.bs,
                                         shuffle=False, num_workers=2)

# get some random training images
images, labels = next(iter(trainloader))
LOG.debug(f"IMAGES = {images.shape}")
LOG.debug(f"LABELS = {labels}")

# show_batch(images.permute(0,2,3,1), labels)

dls = {'train': trainloader, 'val': testloader}
NUM_BATCHES = len(trainset) // CNF.bs
# MODEL
LOG.debug("Loading model...")
net = Net().to(device)
net = nn.SyncBatchNorm.convert_sync_batchnorm(net)
net = DDP(net, device_ids=cuda_ids, find_unused_parameters=True)
# LOG.debug("Loading model summary...")
# summary(net, input_size=(3,32,32))

# LOSS
if LABEL_SMOOTH:
    if SUPERLOSS:
        from jais.train.losses import LabelSmoothingCEWithSuperLoss
        loss_fn = LabelSmoothingCEWithSuperLoss(classes=n_classes, device=device)
    else:
        from jais.train.losses import LabelSmoothingCE
        loss_fn = LabelSmoothingCE(classes=n_classes)
else:
    loss_fn = torch.nn.CrossEntropyLoss()
LOG.info(f"loss function = `{loss_fn.__class__.__name__}`")

# OPTIMIZER AND LR SCHEDULER
optimizer = optim.SGD(net.parameters(),
                      lr=CNF.train.lr,
                      momentum=CNF.train.mom
                      )
optimizer = OptmWithLRSWrapper(optimizer,
                               num_epochs=NUM_EPOCHS,
                               num_iters_per_epochs=NUM_BATCHES,
                               lr_range=(0.01, 1e-5))

metrics = torchmetrics.MetricCollection([
    torchmetrics.Accuracy(num_classes=n_classes, average='macro'),
    torchmetrics.Precision(num_classes=n_classes, average='macro'),
    torchmetrics.Recall(num_classes=n_classes, average='macro'),
    torchmetrics.F1(num_classes=n_classes, average='macro'),
])


def train_one_batch(self, batch):
    # DATA
    inputs, targets = batch
    inputs, targets = inputs.to(self.device), targets.to(self.device)
    inputs, targets_a, targets_b, lam = mixup_data(inputs, targets,
                                                   MIXUP_ALPHA, device)
    # PREDICT
    outputs = self.net(inputs)
    # LOSS
    # loss = self.loss_fn(outputs, targets)
    loss = mixup_criterion(self.loss_fn, outputs, targets_a, targets_b, lam)
    # BACKWARD
    self.optm_fn.zero_grad()
    loss.backward()
    self.optm_fn.step()
    return loss, outputs, targets


def val_one_batch(self, batch):
    # DATA
    inputs, targets = batch
    inputs, targets = inputs.to(self.device), targets.to(self.device)
    # PREDICT
    outputs = self.net(inputs)
    # LOSS
    loss = self.loss_fn(outputs, targets)
    return loss, outputs, targets


Trainer.train_one_batch = train_one_batch
Trainer.val_one_batch = val_one_batch
trainer = Trainer(dls=dls,
                  net=net,
                  loss_fn=loss_fn,
                  optm_fn=optimizer,
                  device=device,
                  metrics=metrics,
                  chkpt_of=[{'val_loss': 'min', 'Accuracy': 'max'}],
                  training_logger_name='wb',
                  training_logs_filename=training_logs_filename,
                  wandb_init_kwargs={'config': CNF},
                  )
trainer.train(NUM_EPOCHS)



from optparse import Option
import os
import math
import time
import torch
from pathlib import Path
from typing import Dict, Optional, Any, Union
from collections import OrderedDict
from torchmetrics import Metric
from wandb import wandb
from jais.train.log import TrainingLogger, RichProgress
from jais.utils import manage_log_files, load_default_configs

CNF, LOG = load_default_configs()


__all__ = ['Trainer']


class Trainer:
    def __init__(self,
                 dls: Dict[str, torch.utils.data.DataLoader],
                 net: torch.nn.Module,
                 loss_fn: torch.nn.Module,
                 optm_fn: torch.nn.Module,
                 device: torch.device,
                 metrics: Optional[Dict[str, Metric]] = None,
                 chkpts_dir: Optional[Union[str, Path]] = None,
                 chkpt_of: Optional[list[Dict[str, str]]] = [{'val_loss': 'min'}],
                 chkpt_prefix: str = "",
                 training_logger_name: Optional[str] = None,
                 training_logs_dir: Optional[Union[str, Path]] = None,
                 training_logs_filename: Optional[Union[str, Path]] = None,
                 keep_n_recent_logs: Optional[int] = 5,
                 wandb_init_kwargs: Optional[dict] = None,
                 wandb_model_versioning: bool = True) -> None:
        """Training and validation handler class

        Args:
            dls: dict of dataloaders
            net: Network / model to train and validate
            loss_fn: criterion instance
            optm_fn: optimizer instance
            device: device to train on.
            metrics: dict of metrics such as accuracy, precision, recall, etc.
            chkpts_dir: folder path to save trained model and checkpoints.
            chkpt_of: List of names and mode pair 
                (`{<metric name>: <min or max>}`) of the metrics to create 
                a checkpoint on best value. e.g. save model weights when 
                val loss converges from best value. 
                The weights will be saved as `baseline_best_val_loss_wts.pt`.
            chkpt_prefix: Name for model weight's save file. e.g. 
                `chkpt_prefix='baseline'` will save the weights with filename
                `baseline_best_val_loss_wts.pt` or 
                `baseline_best_accuracy_wts.pt` (as per `chkpt_of` names).
            training_logger_name: Name of the logger to use out of ('wb', 'tb'). 
                e.g. 'wb' for wandb or 'tb' for tensorboard.
            training_logs_dir: path to the folder to save training logs.
            training_logs_filename: Name of the training logs file.
                This file will save per epoch metrics. 
                This is same as the displayed metrics table.
                [Default is `CNF.log.training_log_filename_prefix@<current time>.csv`]
            wandb_init_kwargs: `wandb.init` function arguments. 
                e.g. {
                    'config': <wandb configurations>
                    'project': <project name>, 
                    'entity': <enitity>,
                }
            wandb_model_versioning: If True, log model weights to wandb.
        """
        self.dls, self.net = dls, net.to(device)
        self.loss_fn, self.optm_fn = loss_fn, optm_fn
        self.device, self.metrics = device, metrics

        # Checkpoints settings
        if chkpts_dir is None:
            self.chkpts_dir = Path.cwd()/'checkpoints'
        else:
            self.chkpts_dir = Path(chkpts_dir)
        self.chkpts_dir.mkdir(exist_ok=True)
        # Save model weights on given metrics
        self.chkpts_handler = {}
        self.chkpt_of = chkpt_of
        for _met_dict in chkpt_of:
            _met_name, _met_mode = list(_met_dict.items())[0]
            if _met_mode == 'min':
                self.chkpts_handler[_met_name] = float('inf')
            elif _met_mode == 'max':
                self.chkpts_handler[_met_name] = 0.
            else:
                _errmsg = "Either pass `min` to save weights when converge "
                _errmsg += "(for losses) or pass `max` to save weights when "
                _errmsg += "metric improves (such as accuracy)"
                raise ValueError(_errmsg)
        self.chkpt_prefix = f"{chkpt_prefix}_" if chkpt_prefix else ''

        # Training logger and settings
        self.training_logger_name = training_logger_name
        self.wandb_model_versioning = wandb_model_versioning
        self.logger = TrainingLogger(logger_name=training_logger_name,
                                     logs_dir=training_logs_dir,
                                     wandb_init_kwargs=wandb_init_kwargs
                                     ) if training_logger_name else None
        # Set training logs folder
        if training_logs_dir is None:
            self.training_logs_dir = Path(CNF.paths.logs_dir)
        else:
            self.training_logs_dir = Path(training_logs_dir)
        # Set training logs filename and extension
        if training_logs_filename is None:
            _path = f"{CNF.log.training_log_filename_prefix}@{time.time()}.csv"
            self.training_logs_filename = _path
        else:
            self.training_logs_filename = f"{training_logs_filename}.csv"
        # Keep only specific number of training log files
        self.keep_n_recent_logs = keep_n_recent_logs
        # Put weights tensor on same device
        if hasattr(self.loss_fn, 'weight') and \
                (self.loss_fn.weight is not None):
            self.loss_fn.weight = self.loss_fn.weight.to(self.device)

    def train_one_batch(self, batch):
        # return loss, outputs, targets
        """
        Either inherit this class to implement new routine
            OR
            def my_train_one_batch_func(self, batch):
                # Routine here
            Trainer.train_one_batch = my_train_one_batch_func
        """
        # DATA
        inputs = batch['inputs'].to(self.device)
        targets = batch['targets'].to(self.device)
        # PREDICT
        outputs = self.net(inputs)
        # LOSS
        loss = self.loss_fn(outputs, targets)
        # BACKWARD
        self.optm_fn.zero_grad()
        loss.backward()
        self.optm_fn.step()
        return loss, outputs, targets

    def val_one_batch(self, batch):
        """
        Either inherit this class to implement new routine
            OR
            def my_val_one_batch_func(self, batch):
                # Routine here
            Trainer.val_one_batch = my_val_one_batch_func
        """
        # DATA
        inputs = batch['inputs'].to(self.device)
        targets = batch['targets'].to(self.device)
        # PREDICT
        outputs = self.net(inputs)
        # LOSS
        loss = self.loss_fn(outputs, targets)
        return loss, outputs, targets

    def train_and_val_one_epoch(self,
                                epoch_num: int,
                                rp: RichProgress = None,
                                max_train_iters: int = None,
                                max_val_iters: int = None) -> Dict[str, Any]:
        """Train and validate one epoch"""
        # Put network in training mode and device
        self.net.train()
        # Train one epoch
        epoch_loss, epoch_outputs, epoch_targets = [], [], []
        for i, batch in enumerate(self.dls['train'], start=1):
            loss, outputs, targets = self.train_one_batch(batch)
            epoch_loss.append(loss)
            epoch_outputs.append(outputs)
            epoch_targets.append(targets)
            rp.update_train_bar()
            if i >= max_train_iters:
                break
        # Compute training metrics
        avg_epoch_loss = torch.tensor(epoch_loss).mean().item()
        # epoch_outputs = torch.cat(epoch_outputs, dim=0).to(self.device)
        # epoch_targets = torch.cat(epoch_targets, dim=0).to(self.device)

        # Validate
        with torch.no_grad():
            self.net.eval()
            val_loss, val_outputs, val_targets = [], [], []
            for i, batch in enumerate(self.dls['val'], start=1):
                loss, outputs, targets = self.val_one_batch(batch)
                val_loss.append(loss)
                val_outputs.append(outputs)
                val_targets.append(targets)
                rp.update_val_bar()
                if i >= max_val_iters:
                    break
            # Compute validation metrics
            avg_val_loss = torch.tensor(val_loss).mean().item()
            val_outputs = torch.cat(val_outputs, dim=0).cpu()
            val_targets = torch.cat(val_targets, dim=0).cpu()
            if self.metrics:
                val_metrics_dict = self.metrics(val_outputs, val_targets)
                val_metrics_dict = {
                    k: v.item() for k, v in val_metrics_dict.items()
                }

        # Update metrics table and logger
        epoch_logs = OrderedDict({
            'epoch_num': epoch_num,
            'train_loss': avg_epoch_loss,
            'val_loss': avg_val_loss,
        })
        if self.metrics:
            epoch_logs.update(val_metrics_dict)

        # Save checkpoints when metric improves
        self._create_checkpoint(epoch_logs)
        return epoch_logs

    def _create_checkpoint(self, epoch_logs) -> None:
        """Save checkpoints when metric improves"""
        for _met_dict in self.chkpt_of:
            _met_name, _met_mode = list(_met_dict.items())[0]
            if _met_name not in epoch_logs.keys():
                _errmsg = f"Metric `{_met_name}` must match names in given" 
                _errmsg += "metrics. Available metrics are "
                _errmsg += f"{self.metrics.keys()}."
                raise KeyError(_errmsg)
            savename = f"{self.chkpt_prefix}best_{_met_name}_wts.pt"
            savepath = self.chkpts_dir/ savename
            _msg  = f"`{_met_name}` improved from "
            _msg += f"{self.chkpts_handler[_met_name]:.5f} to "
            _msg += f"{epoch_logs[_met_name]:.5f}. "
            _msg += f"Checkpoint saved @ `{savepath}`."
            if _met_mode == 'min':
                if epoch_logs[_met_name] < self.chkpts_handler[_met_name]:
                    torch.save(self.net.state_dict(), savepath)
                    self.chkpts_handler[_met_name] = epoch_logs[_met_name]
                    LOG.info(_msg)
            else:
                if epoch_logs[_met_name] > self.chkpts_handler[_met_name]:
                    torch.save(self.net.state_dict(), savepath)
                    self.chkpts_handler[_met_name] = epoch_logs[_met_name]
                    LOG.info(_msg)
        
    def train(self,
              n_epochs: int,
              max_train_iters: int = None,
              max_val_iters: int = None,
              optm_fn=None):
        if optm_fn:  # change optimizer
            self.optm_fn = optm_fn
        self.table_columns = ['Epoch', 'TrainLoss']
        if max_train_iters is None:
            max_train_iters = math.ceil(
                len(self.dls['train'].dataset) / self.dls['train'].batch_size
            )
        if (max_val_iters is None) and ('val' in self.dls.keys()):
            max_val_iters = math.ceil(
                len(self.dls['val'].dataset) / self.dls['val'].batch_size
            )
        if 'val' in self.dls.keys():
            self.table_columns.append('ValLoss')
        # Add metrics names as table columns
        self.table_columns += [
            f"Val{k.capitalize()}" for k in self.metrics.keys()
        ]

        with RichProgress(columns=self.table_columns,
                          n_epochs=n_epochs,
                          n_train_iters=max_train_iters,
                          n_val_iters=max_val_iters) as rp:
            for epoch_num in range(1, n_epochs + 1):

                epoch_logs = self.train_and_val_one_epoch(
                    epoch_num=epoch_num,
                    rp=rp,
                    max_train_iters=max_train_iters,
                    max_val_iters=max_val_iters
                )
                if self.logger:
                    self.logger.log(epoch_logs)
                rp.update_epoch(list(epoch_logs.values()))

        # Save last epoch model weights
        savepath = self.chkpts_dir/f'{self.chkpt_prefix}last_epoch_wts.pt'
        torch.save(self.net.state_dict(), savepath)

        # Log model to wandb, if given
        if (self.training_logger_name in ['wb', 'wandb']) and \
            (self.wandb_model_versioning):
            trained_model_artifact = wandb.Artifact(
                f'TrainedModel-{self.net.__class__.__name__}',
                type='model')
            trained_model_artifact.add_dir(self.chkpts_dir)
            wandb.log_artifact(trained_model_artifact)

        # Save logs
        manage_log_files(logs_dir=self.training_logs_dir,
                         keep_n_recent_logs=self.keep_n_recent_logs,
                         file_ext='.csv')
        rp.to_csv(self.training_logs_dir/self.training_logs_filename)



# Package Info
NAME = 'JAIS'
VERSION = '0.0.1.13'
DESCRIPTION = "Just Artificial Intelligence Snippets (JAIS)"

# Basic settings
from pathlib import Path
ROOT_DIR = Path(__file__).parent  # Package path
PROJ_DIR = ROOT_DIR.parent  # Project path


#!/usr/bin/env python
from setuptools import setup, find_packages
from jais.__init__ import NAME, VERSION, DESCRIPTION
from jais.utils.fileloader import save_json

NAME = NAME.lower()

with open("requirements.txt") as rf:
    REQUIREMENTS = [l.strip() for l in rf.read().split("\n")]

with open('README.md') as rf:
    LONG_DESCRIPTION = rf.read()

save_json({},f'{NAME}/configs/{NAME}_settings.json')

setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      author='Jitender Singh Virk',
      author_email='krivsj@gmail.com',
      url='https://github.com/VirkSaab/JAIS',
      packages=find_packages(exclude=['ez_setup', 'tests*', ".github"]),
      # package_data={'jais': ['data/*.dat']},
      include_package_data=True,
      install_requires=REQUIREMENTS,
      setup_requires=['wheel'],
      entry_points=f"""
        [console_scripts]
        {NAME}={NAME}.main:cli
      """)



import os
from typing import Union, Any, Dict
from collections import OrderedDict
from pathlib import Path
from easydict import EasyDict as edict
from ruamel.yaml import YAML, yaml_object
from jais.__init__ import ROOT_DIR

__all__ = [
    # JSON
    'load_json', 'save_json',

    # YAML
    'load_yaml', 'save_yaml',
]


# ----------------------------------------> JSON :
def load_json(path: Union[str, Path]) -> edict:
    """Load .json file from given path"""
    import json
    with open(path) as f:
        return edict(json.load(f))


def save_json(object: Any, path: Union[str, Path]) -> None:
    """Save .json file to given path"""
    import json
    path = Path(path).resolve()
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(str(path), 'w') as f:
        json.dump(object, f)


# ----------------------------------------> YAML :
# * When using the decorator, which takes the YAML() instance as a parameter,
# * the yaml = YAML() line needs to be moved up in the file -- *yaml docs*
yaml = YAML(typ='safe', pure=True)
yaml.default_flow_style = False
yaml.indent(mapping=2, sequence=4, offset=2)


@yaml_object(yaml)
class JoinPath:
    """Custom tag `!join` loader class to join strings for yaml file."""

    yaml_tag = u'!joinpath'

    def __init__(self, joined_string):
        self.joined_string = joined_string

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag,
                                            u'{.joined_string}'.format(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        seq = constructor.construct_sequence(node)
        fullpath = Path(os.path.join(*[str(i) for i in seq])).resolve()
        return str(fullpath)


@yaml_object(yaml)
class RootDirSetter:
    """Custom tag `!rootdir` loader class for yaml file."""

    yaml_tag = u'!rootdir'

    def __init__(self, path):
        self.path = path

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag,
                                            u'{.path}'.format(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        return str(ROOT_DIR)


@yaml_object(yaml)
class CWDSetter:
    """Custom tag `!cwd` loader class for yaml file."""

    yaml_tag = u'!cwd'  # Current Working Directory

    def __init__(self, path):
        self.path = path

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag,
                                            u'{.path}'.format(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        return load_json(ROOT_DIR/'configs/jais_settings.json').JAIS_CWD


@yaml_object(yaml)
class HomeDirSetter:
    """Custom tag `!homedir` loader class for yaml file."""

    yaml_tag = u'!homedir'

    def __init__(self, path):
        self.path = path

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag,
                                            u'{.path}'.format(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        return str(Path.home())


def load_yaml(path: Union[str, Path], pure: bool = False) -> dict:
    """config.yaml file loader.

    This function converts the config.yaml file to `dict` object.

    Args:
        path: .yaml configuration filepath
        pure: If True, just load the .yaml without converting to EasyDict
            and exclude extra info.

    Returns:
        `dict` object containing configuration parameters.

    Example:
        .. code-block:: python

            config = load_yaml("../config.yaml")
            print(config["project_name"])
    """

    path = str(Path(path).absolute().resolve())
    # * Load config file
    with open(path) as file:
        config = yaml.load(file)

    if pure == False:  # Add extra features
        # Convert dict to easydict
        config = edict(config)
    return config


def save_yaml(config_dict: Dict, path: Union[str, Path] = None,
                saveas_ordered: bool = True, force_overwrite: bool = False,
                file_extension: str = 'yaml') -> None:
    """save `dict` config parameters to `.yaml` file

    Args:
        config_dict: parameters to save.
        path: path/to/save/auto_config.yaml.
        saveas_ordered: save as collections.OrderedDict on Python 3 and 
            `!!omap` is generated for these types.
        file_extension: default `.yaml`

    Returns: 
        nothing

    Example:
        .. code-block:: python

            config = {"Example": 10}
            save_config(config, "../auto_config.yaml")
    """

    if path is None:  # set default path if not given
        path = config_dict.paths.output_dir
        path = f"{path}/modified_config.{file_extension}"
    else:
        path = str(Path(path).absolute().resolve())
        # Check if the path given is the original config's path
        if (config_dict.original_config_filepath == path) and \
           (not force_overwrite):
            msg = f"""
            Error while saving config file @ {path}.
            Cannot overwrite the original config file
            Choose different save location.
            """
            raise ValueError(msg)

    # converting easydict format to default dict because
    # YAML does not processes EasyDict format well.
    cleaned_dict = {
        k: edict_to_dict_converter(v)
        for k, v in config_dict.items()
    }
    # print("cleaned_dict =", cleaned_dict)

    # Fix order
    if saveas_ordered:
        cleaned_dict = OrderedDict(cleaned_dict)

    # Save the file to given location
    with open(path, 'w') as file:
        yaml.dump(cleaned_dict, file)

    print(f"config saved @ {path}")


def edict_to_dict_converter(x: Union[Dict, Any]) -> Union[Dict, Any]:
    """Recursive function to convert given dictionary's datatype
    from edict to dict.

    Args:
        x (dict or other): nested dictionary

    Returns: x (same x but default dict type)
    """
    if not isinstance(x, dict):
        return x

    # Recursion for nested dicts
    ndict = {}
    for k, v in x.items():
        v1 = edict_to_dict_converter(v)
        ndict[k] = v1
    return ndict
